% This file is a part of the CaosDB Project.
%
% Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
% Copyright (C) 2021 Daniel Hornung <d.hornung@indiscale.com>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <https://www.gnu.org/licenses/>.

% -*- texinfo -*-
% @deftypefn {Class} {} Caosdb ()
% This is the main class of the CaosDB client for Octave.
%
% Print usage help and return:
% @example
% caosdb --help;
% @end example
%
% Print the version and return:
% @example
% caosdb --version;
% @end example
%
% Test the default connection, print the server's version and return.
% @example
% caosdb --test-connection;
% @end example
%
% @end deftypefn

%!demo
%! ## Print Help:
%! caosdb --help;
classdef Caosdb < handle
  properties
    connection = ""             % Will use the default connection
  end

  methods

    function obj = Caosdb(connection)
      if nargin == 0
        connection = "";
      end
      obj.connection = connection;
    end

    %%
    %% info
    %%
    function res = info(obj)
      % disp(["connection: >", obj.connection, "<"]);
      try
        info_result = maox_info(obj.connection);
        res = info_result;
      catch
        % disp("some error!");
        % disp(lasterror());
        rethrow(lasterror());
      end
    end

    %%
    % Retrieve entities by IDs
    %
    % entities = Caosdb.retrieve_by_id("my_id", {"more", "ids"})
    %
    % Parameters
    % ----------
    %
    % ids: string or cell array of strings
    %  The ID(s) of the entity (entities) to be retrieved.
    %
    % Returns
    % -------
    % entities : cell array
    %  The retrieved entities.
    function entities = retrieve_by_id(obj, ids, varargin)
      % Ensure that IDS is a string cell array
      if ischar(ids)
        ids = {ids};
      else
        assert(iscellstr(ids), ...
               "maox:InvalidArgument", ...
               "IDS must be a string or string cell array, was:\n%s", ...
               disp(ids));
      end
      % Make one big cell array of potential strings
      for argin = varargin
        if ischar(argin)
          ids(end + 1) = argin;
        else
          assert(iscellstr(argin), ...
                 "maox:InvalidArgument", ...
                 "arguments must be a string (cell array), but this argument was found:\n%s", ...
                 disp(argin));
          ids(end + 1:end + numel(argin)) = argin(:);
        end
      end

      % disp("ids:");
      % disp(ids);
      try
        collection = maox_retrieve(obj.connection, ids);
        entities = maox_convert_collection(collection);
      catch
        % disp("error handling in Caosdb.m");
        % disp(lasterror());
        rethrow(lasterror());
      end
    end

    %%
    % Execute a query
    %
    % entities = Caosdb.query("FIND Record Foo WITH bar=baz")
    %
    % Parameters
    % ----------
    %
    % query_str: string
    %  The query to be executed.
    %
    % Returns
    % -------
    % entities : cell array
    %  The retrieved entities.
    function entities = query(obj, query_str)
      % Ensure that QUERY is a string.
      assert(ischar(query_str), "maox:InvalidArgument", "QUERY must be a string, was:\n%s", ...
             disp(query_str));
      assert(nargin == 2, "maox:InvalidArgument", "This method accepts exactly 1 argument.");
      % disp("query:");
      % disp(query_str);
      try
        collection = maox_query(obj.connection, query_str);
        entities = maox_convert_collection(collection);
      catch
        % disp("error handling in Caosdb.m");
        % disp(lasterror());
        rethrow(lasterror());
      end
    end

  end
end
