% This file is a part of the CaosDB Project.
%
% Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
% Copyright (C) 2021 Daniel Hornung <d.hornung@indiscale.com>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <https://www.gnu.org/licenses/>.

classdef Entity < handle

  properties
    role        % One of "RecordType", "Record", "Property"
    id
    versionId
    name
    description
    datatype
    unit
    value
  end
  properties (Access = private)
    parents_     % Cell array of Parent objects.Struct array with the following fields:
    %  - id
    %  - name
    %  - description
    properties_  % Cell array of Property objects.Struct array with the following fields:
    %  - id
    %  - name
    %  - description
    %  - importance
    %  - value
    %  - unit
    %  - datatype
    errors_      % Cell array of Message objects.
    warnings_    % Like messages.
    infos_       % Like messages.
  end

  methods

    % Constructor.
    % The structure of the DATA parameter follows the convention outline in `maoxdb.hpp`.
    function obj = Entity(data)
      narginchk(1, 1);
      obj.assign_scalars(data);
      obj.assign_parents(data);
      obj.assign_properties(data);
      obj.errors_ = obj.create_messages(data, "errors");
      obj.warnings_ = obj.create_messages(data, "warnings");
      obj.infos_ = obj.create_messages(data, "infos");
    end

    % Getters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function result = get_parents(obj)
      result = obj.parents_;
    end

    function result = get_properties(obj)
      result = obj.properties_;
    end

    function result = get_errors(obj)
      result = obj.errors_;
    end

    function result = get_warnings(obj)
      result = obj.warnings_;
    end

    function result = get_infos(obj)
      result = obj.infos_;
    end

    % Information %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function result = has_errors(obj)
      result = not(isempty(obj.errors_));
    end

    function result = has_warnings(obj)
      result = not(isempty(obj.warnings_));
    end

    function result = has_infos(obj)
      result = not(isempty(obj.infos_));
    end

  end

  methods (Access = private)

    % Only the simple metadata fields.
    function assign_scalars(obj, data)
      obj.role = data.role;
      obj.id = data.id;
      obj.versionId = data.versionId;
      obj.name = data.name;
      obj.description = data.description;
      obj.datatype = data.datatype;
      obj.unit = data.unit;
      obj.value = data.value;
    end

    % Assign the parents
    function assign_parents(obj, data)
      obj.parents_ = cell();
      for parent = data.parents
        obj.parents_{end + 1} = Parent(parent);
      end
    end

    % Assign the properties
    function assign_properties(obj, data)
      obj.properties_ = cell();
      for property = getfield(data, "properties")
        obj.properties_{end + 1} = Property(property);
      end
    end

    % Create the errors, warnings, infos
    function result = create_messages(~, data, level)
      result = cell();
      for message = getfield(data, level)
        result{end + 1} = Message(message);
      end
    end

  end

end
