#ifndef MAOXDB_H // NOLINT
#define MAOXDB_H

#include "caosdb/exceptions.h"
#include "caosdb/status_code.h"
#include "caosdb/transaction.h"
#include "caosdb/transaction_status.h"
#include "mex.h"
#include <boost/algorithm/string.hpp>
#include <stdexcept>
#include <string>

/**
 * Mostly utility functions for MEX files interacting with libcaosdb.
 *
 * \b{Data exchange}
 *
 * Exchanging data between the Octave side wrapper and this C++ library is done
 * via Octave struct arrays, each struct corresponds to one CaosDB Entity.
 *
 * The Entity structs have the following structure (not all keys may exist for
 * all roles):
 *
 * - role:        One of "RecordType", "Record", "Property"
 * - id
 * - versionId
 * - name
 * - description
 * - datatype
 * - unit
 * - value:       A string representation of the value.
 * - parents:     Struct array with the following fields:
 *                - id
 *                - name
 *                - description
 * - properties:  Struct array with the following fields:
 *                - id
 *                - name
 *                - description
 *                - importance
 *                - value
 *                - unit
 *                - datatype
 * - errors:      Struct array with the following fields:
 *                - code (as INT64)
 *                - description
 * - warnings:    Like messages.
 * - infos:       Like messages.
 *
 */

// Macros /////////////////////////////////////////////////////////////////////

// Utility functions //////////////////////////////////////////////////////////
namespace maoxdb {

using std::string;

template <typename T> auto mxScalar(T value, mxClassID class_id) -> mxArray * {
  mxArray *array = mxCreateNumericMatrix(1, 1, class_id, mxREAL);
  auto *data = static_cast<T *>(mxGetData(array));
  data[0] = value;
  return array;
}

// Convenience shortcut functions

inline auto mxScalarUINT64(UINT64_T value) -> mxArray * {
  return mxScalar<UINT64_T>(value, mxUINT64_CLASS);
}
inline auto mxScalarINT64(INT64_T value) -> mxArray * {
  return mxScalar<INT64_T>(value, mxINT64_CLASS);
}
inline auto mxScalarDOUBLE(double value) -> mxArray * {
  return mxScalar<double>(value, mxDOUBLE_CLASS);
}
inline auto mxScalarLOGICAL(bool value) -> mxArray * {
  return mxScalar<UINT8_T>(static_cast<unsigned char>(value), mxLOGICAL_CLASS);
}
inline auto mxEmptyDOUBLE() -> mxArray * {
  mxArray *array = mxCreateNumericMatrix(0, 0, mxDOUBLE_CLASS, mxREAL);
  return array;
}
inline auto mxEmptySTRING() -> mxArray * {
  mxArray *array = mxCreateString("");
  return array;
}
inline auto mxEmptySTRUCT() -> mxArray * {
  mxArray *array = mxCreateStructArray(0, (mwSize const *)nullptr, 0, (char const **)nullptr);
  return array;
}

/**
 * @brief Convert the string-typed value in an Entity-like object to a mxArray.
 *
 * @details For the existing datatypes, see
 * https://docs.indiscale.com/caosdb-server/specification/Datatype.html
 */
template <class T> auto mxScalarFromStringValue(const T &entity) -> mxArray * {
  mxArray *result;
  auto dt = entity.GetDatatype();
  auto value = entity.GetValue();
  // std::cout << "Datatype/Value: " << dt << ": " << value << std::endl;
  if (dt.empty()) {
    result = mxEmptySTRING();
  } else if (dt == "TEXT") {
    result = mxCreateString(value.c_str());
  } else if (dt == "BOOLEAN") {
    if (boost::to_upper_copy(value) == "TRUE") {
      result = mxScalarLOGICAL(1);
    } else {
      if (boost::to_upper_copy(value) != "FALSE") {
        mxAssert(boost::to_upper_copy(value) == "FALSE",
                 std::string("Value >") + value + "< is neither FALSE nor TRUE.");
        throw std::invalid_argument(std::string("Value >") + value +
                                    "< is neither FALSE nor TRUE.");
      }
      result = mxScalarLOGICAL(0);
    }
  } else if (dt == "INTEGER") {
    result = mxScalarINT64(stol(value));
  } else if (dt == "DOUBLE") {
    result = mxScalarDOUBLE(stod(value));
  } else if (dt == "FILE") {
    result = mxCreateString(value.c_str());
  } else if (dt == "REFERENCE") {
    result = mxCreateString(value.c_str());
  } else if (dt == "DATETIME") {
    result = mxCreateString(value.c_str());
  } else if (dt.substr(0, 4) == "LIST") {
    mexErrMsgTxt("List values are not implemented yet.");
  } else {
    std::cout << "Unknown datatype, probably a reference: " << dt << ": " << value << std::endl;
    result = mxCreateString(value.c_str());
  }
  mxAssert(result, "Scalar value must have some result now.");
  return result;
}

/**
 * @brief      Extract a std::string from a char array
 *
 * @details    Handles the length check
 *
 * @param      array The char array from which the string shall be extracted.
 *
 * @return     The std::string.
 */
auto mxGetStdString(const mxArray *array) -> std::string;

// Exception and error handling ///////////////////////////////////////////////

/**
 * @brief      Generate a mex error message from a TransactionStatus.
 *
 * @details    Preprocess the transaction status of a transaction, to create the
 * necessary information for raising an Octave error.
 *
 * @param      trans
 *  The transaction of which to handle the status.
 *
 * @return     std::pair<std::string, std::string>
 *  The message ID and description of any error.  If the transaction status is
 *  SUCCESS, or some kind of "still running", both strings are empty.
 */
auto transactionStatusToMessage(const caosdb::transaction::Transaction *const trans)
  -> std::pair<string, string>;

/**
 * @brief Throw an Octave error if the transaction did not finish properly.
 *
 * @details On error, this function will leave this C++ context and fall back to
 * Octave.
 *
 * @param   trans
 *  The transaction to check.
 */
void throwOctExceptionIfError(const caosdb::transaction::Transaction *const trans);

/**
 * @brief Handle a CaosDB Exception and transform into error code and message.
 *
 * @details This function does all the processing to transform the information
 * in an Exception object into a form that can conveniently be handled by Octave
 * via the mexErrMsgIdAndTxt function.
 *
 * @param exc The Exception object.
 *
 * @return pair<string msgId, string msgText> The message ID and message text
 * which can be passed to Octave.
 *
 * @note Additional utility functions should be easy to implement when the
 * libcaosdb error handling changes.
 */
auto exceptionToMessage(const caosdb::exceptions::Exception &exc) -> std::pair<string, string>;

/**
 * @brief Throw a CaosDB Exception inside Octave.
 *
 * @details This function does all the processing to transform the information
 * in an Exception object into a form that is accepted by Octave's error
 * handling.d
 *
 * Internally, this function uses the exceptionToMessage function.
 *
 * @param exc The Exception object.
 */
void throwOctException(const caosdb::exceptions::Exception &exc);

// Entity handling ////////////////////////////////////////////////////////////

/**
 * @brief Convert a ResultSet to a struct mexArray.
 *
 * @return A 1xN struct array, with Entity structs as values.
 */
auto mxFromResultSet(const caosdb::transaction::ResultSet &resultSet) -> mxArray *;

/**
 * @brief Convert an Entity from libcaosdb to Octave struct mexArray.
 *
 * @return A 1x1 struct array whose entry follows the maoxdb Entity convention
 * outlined above.
 */
auto mxFromCaosDBEntity(const caosdb::entity::Entity &entity) -> mxArray *;

// Utility functions //////////////////////////////////////////////////////////

/**
 * @brief      Merges a number of scalar mex structs into a 1xN struct.
 *
 * @details    The content (the mxArrays behind the input structs values) is
 * duplicated.
 *
 * @param      structs: Must all have the same fields.
 *
 * @return     A 1xN struct array.
 */
auto mxMergeScalarStructs(const std::vector<const mxArray *> &structs) -> mxArray *;

} // namespace maoxdb

#endif /* MAOXDB_H */
