#include "caosdb/connection.h" // for Connection, ConnectionManager
#include "caosdb/constants.h"  // for LIBCAOSDB_VERSION_MAJOR, LIBCAOSDB_VE...
#include "caosdb/exceptions.h" // for all error handling
#include "caosdb/info.h"       // for VersionInfo
#include "maoxdb.hpp"          // caosDB utils for mex files
#include "mex.h"               // for mxArray, mexFunction
#include <cstring>             // for strcmp
#include <memory>              // for unique_ptr, __shared_ptr_access, shar...
#include <string>              // for allocator, char_traits, operator+

using caosdb::connection::Connection;
using caosdb::connection::ConnectionManager;
using caosdb::transaction::Transaction;
using caosdb::transaction::TransactionStatus;
using std::string;

auto info(const string &connection_name) -> mxArray *;

/**
 * The implementation of the info retrieval.
 */
auto info(string const &connection_name) -> mxArray * {
  std::shared_ptr<Connection> connection = nullptr;
  if (connection_name.empty()) {
    connection = ConnectionManager::GetDefaultConnection();
  } else {
    connection = ConnectionManager::GetConnection(connection_name);
  }
  const auto &version_info = connection->RetrieveVersionInfo();
  const char *keys[] = {"major", "minor", "patch", "pre_release", // NOLINT
                        "build"};
  std::array<mwSize, 2> dims = {1, 1};
  mxArray *info_struct = mxCreateStructArray(2, dims.data(), 5, keys); // NOLINT

  mxSetField(info_struct, 0, "major", maoxdb::mxScalarUINT64(version_info.GetMajor()));
  mxSetField(info_struct, 0, "minor", maoxdb::mxScalarUINT64(version_info.GetMinor()));
  mxSetField(info_struct, 0, "patch", maoxdb::mxScalarUINT64(version_info.GetPatch()));
  mxSetField(info_struct, 0, "pre_release", mxCreateString(version_info.GetPreRelease().c_str()));
  mxSetField(info_struct, 0, "build", mxCreateString(version_info.GetBuild().c_str()));
  return info_struct;
}

/**
 * @brief      Retrieve one or more entities.
 *
 * @details    This function returns the entities as a cell array.
 *
 * @param connection_name A string with the connection name. May be omitted or
 * an empty string.
 *
 * @param IDs One or more Entity IDs, encapsulated in a string cell array.
 *
 * @return collection A struct with the entities.
 */
void mexFunction(int /*nlhs*/, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

  string conn_name;
  std::vector<string> ids;
  try {
    if (nrhs < 2) {
      mexErrMsgIdAndTxt("maox:InsufficientArguments",
                        "Need 2 arguments: connection_name and a string cell array with IDs.");
    }
    if (nrhs > 2) {
      mexErrMsgIdAndTxt("maox:TooManyArguments",
                        "Need 2 arguments: connection_name and a string cell array with IDs.");
    }
    if (!mxIsCell(prhs[1])) {
      mexErrMsgIdAndTxt("maox:InvalidArgument", "The second argument must be a string cell array.");
    }
    if (mxGetNumberOfElements(prhs[0]) > 0) {
      conn_name = mxArrayToString(prhs[0]);
    }

    // Collect the IDs
    for (auto i = 0; i < mxGetNumberOfElements(prhs[1]); ++i) {
      try {
        auto id = maoxdb::mxGetStdString(mxGetCell(prhs[1], i));
        ids.emplace_back(id);
      } catch (const std::invalid_argument &exc) {
        mexErrMsgIdAndTxt("maox:InvalidArgument",
                          "The second argument must be a string cell array.");
      }
    }
  } catch (...) {
    mexErrMsgIdAndTxt("maox:ArgumentHandling", "Error while handling the arguments.");
  }
  auto n_ids = ids.size();

  // Retrieve the Entities
  std::shared_ptr<Connection> connection = nullptr;
  if (conn_name.empty()) {
    connection = ConnectionManager::GetDefaultConnection();
  } else {
    connection = ConnectionManager::GetConnection(conn_name);
  }
  auto transaction = connection->CreateTransaction();
  for (const auto &id : ids) {
    // std::cout << "RetrieveById(" << id << ")" << std::endl;
    transaction->RetrieveById(id);
  }
  transaction->ExecuteAsynchronously();
  auto t_stat = transaction->WaitForIt();
  maoxdb::throwOctExceptionIfError(transaction.get());
  // Status must be OK or GENERIC_TRANSACTION_ERROR now.
  const auto &results = transaction->GetResultSet();
  // std::cout << "size: " << results.Size() << std::endl;

  auto *mxResults = maoxdb::mxFromResultSet(results);

  plhs[0] = mxDuplicateArray(mxResults);
}
