/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2021 Daniel Hornung <d.hornung@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "caosdb/exceptions.h"
#include "maoxdb.hpp"
#include "mex.h"
// #include "mexproto.h"
#include <boost/lexical_cast.hpp>
#include <gtest/gtest.h>
#include <charconv>
#include <limits>
#include <memory>
#include <string>
#include <type_traits>
#include <vector>

using std::string;

namespace maoxdb {

///////////////////////////////////////////////////////////////////////////////
//                              Helper functions                             //
///////////////////////////////////////////////////////////////////////////////

// Helper implementation ...
template <typename T> void test_value_impl(const mxArray *mxValue, const T &value) {
  EXPECT_EQ(mxGetNumberOfElements(mxValue), 1);
  if (value != value) { // equality test does not make sense for NaN
    EXPECT_NE(*((T *)mxGetData(mxValue)), *((T *)mxGetData(mxValue)));
  } else {
    EXPECT_EQ(*((T *)mxGetData(mxValue)), value);
  }
}
// ... and specializations.
template <> // string
void test_value_impl(const mxArray *mxValue, const string &value) {
  EXPECT_EQ(mxGetNumberOfElements(mxValue), value.size());
  EXPECT_EQ(mxGetStdString(mxValue), value);
}

/**
 * Can be instantiated for e.g. Entity and Property.
 *
 * This expects T to be either "scalar" types, or std::string.
 */
template <typename EntType, typename T>
void test_type_value(const string &datatype, const std::vector<T> &values,
                     const std::vector<string> &validStrings,
                     const std::vector<string> &invalidStrings,
                     const std::vector<string> &expectedErrors) {
  EXPECT_EQ(values.size(), validStrings.size());
  EXPECT_EQ(invalidStrings.size(), expectedErrors.size());
  auto valueEntity = EntType();
  valueEntity.SetDatatype(datatype);

  for (size_t i = 0; i < values.size(); ++i) {
    auto value = values[i];
    auto valueString = validStrings[i];
    // std::cout << "---- " << value << " ----" << std::endl;
    valueEntity.SetValue(valueString);
    auto mxValue = mxScalarFromStringValue(valueEntity);
    test_value_impl<T>(mxValue, value);
  }
  for (size_t i = 0; i < invalidStrings.size(); ++i) {
    auto invalidString = invalidStrings[i];
    auto expectedError = expectedErrors[i];
    try {
      valueEntity.SetValue(invalidString);
      auto mxValue = mxScalarFromStringValue(valueEntity);
      ADD_FAILURE() << "Expected error for value '" << invalidString << "'.";
    } catch (std::exception const &err) {
      EXPECT_EQ(err.what(), expectedError);
    } catch (...) {
      ADD_FAILURE() << "Expected std::exception for value '" << invalidString << "'.";
    }
  }
}

/**
 * Generate a vector with useful values for testing.
 */
template <typename T> auto generateValues() -> std::vector<T> {
  std::vector<T> values = {
    0,
    1,
    std::numeric_limits<T>::max(),
    std::numeric_limits<T>::min(),
    std::numeric_limits<T>::lowest(),
    std::numeric_limits<T>::epsilon() // 0 for integers, but who cares?
  };
  return values;
}

/**
 * Convenience wrapper around std::to_chars
 */
template <typename T> auto toStr(T value) -> std::string {
  char *chars = reinterpret_cast<char *>(std::malloc(50 * sizeof(char)));
  std::to_chars(chars, chars + 50, value);
  auto result = string(chars);
  std::free(chars);
  return result;
}

// to_chars is not widely supported for floats yet.
template <> auto toStr(double value) -> std::string {
  auto result = boost::lexical_cast<std::string>(value);
  return result;
}

///////////////////////////////////////////////////////////////////////////////
//                                Actual tests                               //
///////////////////////////////////////////////////////////////////////////////

/**
 * Test if value conversion works
 */
TEST(caosdb_conversion, value_TEXT) {
  auto values = std::vector<string>{"", "23", "foo", "CaosDB", "\n\n\n\a\n</>"};
  auto validStrings = std::vector<string>(values);
  auto invalidStrings = std::vector<string>();
  auto expectedErrors = std::vector<string>();
  for (auto datatype : std::vector<string>{"TEXT", "REFERENCE", "DATETIME", "custom name"}) {
    test_type_value<caosdb::entity::Entity, string>(datatype, values, validStrings, // NOSTYLE
                                                    invalidStrings, expectedErrors);
    test_type_value<caosdb::entity::Property, string>(datatype, values, validStrings,
                                                      invalidStrings, expectedErrors);
  }
}

TEST(caosdb_conversion, value_BOOLEAN) {
  auto values = std::vector<bool>{true, true, true, false, false, false};
  auto validStrings = std::vector<string>{"true", "True", "TRUE", "false", "fAlSe", "FALSE"};
  auto invalidStrings = std::vector<string>{"no", "yes", "MAYBE"};
  auto expectedErrors = std::vector<string>{"Value >no< is neither FALSE nor TRUE.",
                                            "Value >yes< is neither FALSE nor TRUE.",
                                            "Value >MAYBE< is neither FALSE nor TRUE."};
  test_type_value<caosdb::entity::Entity, bool>("BOOLEAN", values, validStrings, // NOSTYLE
                                                invalidStrings, expectedErrors);
  test_type_value<caosdb::entity::Property, bool>("BOOLEAN", values, validStrings, // NOSTYLE
                                                  invalidStrings, expectedErrors);
}

TEST(caosdb_conversion, value_INTEGER) {
  auto values = generateValues<long>();
  auto validStrings = std::vector<string>(values.size());
  std::transform(values.begin(), values.end(), validStrings.begin(),
                 [](long value) -> string { return std::to_string(value); });
  auto invalidStrings = std::vector<string>{"NaN", "zero", "many"};
  auto expectedErrors = std::vector<string>{"stol", "stol", "stol"};
  test_type_value<caosdb::entity::Entity, long>("INTEGER", values, validStrings, // NOSTYLE
                                                invalidStrings, expectedErrors);
  test_type_value<caosdb::entity::Property, long>("INTEGER", values, validStrings, // NOSTYLE
                                                  invalidStrings, expectedErrors);
}

TEST(caosdb_conversion, value_DOUBLE) {
  auto values = generateValues<double>();
  values.push_back(std::numeric_limits<double>::quiet_NaN());
  values.push_back(std::numeric_limits<double>::infinity());
  values.push_back(-std::numeric_limits<double>::infinity());
  auto validStrings = std::vector<string>(values.size());
  std::transform(values.begin(), values.end(), validStrings.begin(),
                 [](double value) -> string { return toStr<double>(value); });
  // for (size_t i = 0; i < values.size(); ++i) {
  //   std::cout << values[i] << " -> " << validStrings[i] << std::endl;
  // }
  auto invalidStrings = std::vector<string>{"dunno", "zero", "many"};
  auto expectedErrors = std::vector<string>{"stod", "stod", "stod"};
  test_type_value<caosdb::entity::Entity, double>("DOUBLE", values, validStrings, // NOSTYLE
                                                  invalidStrings, expectedErrors);
  test_type_value<caosdb::entity::Property, double>("DOUBLE", values, validStrings, // NOSTYLE
                                                    invalidStrings, expectedErrors);
}

/*
 * The following are treated just like TEXT, at least for the moment:
 * - REFERENCE
 * - DATETIME
 * - References with other names
 *
 * No testing exists yet for:
 * - FILE
 */

} // namespace maoxdb
